import time
from dataclasses import dataclass

import numpy as np
from .welford import welford
from .dense import ttcDense, ttcDense_int
from .fxpcs_sparse import sparsify, twoTime, twoTimeSymmNorm

class WelfordAlg:

    def __init__(self, size, dtype=np.float32):
        self.dtype = dtype
        self.count   = np.zeros(size, dtype=dtype)
        self.avg     = np.zeros(size, dtype=dtype)
        self.var     = np.zeros(size, dtype=dtype)
        
    def update(self, x, mask=None, alpha=None):
        x = x.astype(self.dtype)
        cond = ~mask & (self.count == 0)
        self.avg = np.where(~mask, self.avg, x)
        self.var = np.where(~mask, self.var, 0.0)

        alph_arr = np.where(self.count, 1/self.count, 0.0)
        if alpha is not None:
            alph_arr[:] = np.where(~mask, alpha, 0.0)
        welford(x, self.avg, self.var, alph_arr)
        self.count = np.where(~mask, self.count + 1, self.count)

### MATH ON TRAINS 

def do_stats(iterator, *args):
    
    tids = list()
    
    tid, img, cmsk = iterator.__next__()
    tids.append(tid)
    
    wf = WelfordAlg(img.shape)

    _msk = (cmsk > 0).astype(np.bool_)
    wf.update(img, _msk)

    for tid, img, cmsk in iterator:
        tids.append(tid)
        _msk = (cmsk > 0).astype(np.bool_)
        wf.update(img, _msk) 
    
        #print(tid, wf.count.mean()) 
        
    return tids, wf.avg, wf.var, wf.count

##############
# DENSE TTC
##############

def do_dense(iterator, m):
    
#   all_ttcs_on = list()
#   all_ttcs_off= list()

    tids = list()
    
    tid, img, cmsk = iterator.__next__()
    tids.append(tid)
    j = 0

    prevImg = np.ones_like(img, dtype=img.dtype)
    prevMsk = np.ones_like(m, dtype=m.dtype)

    _msk = (m & ~np.any(cmsk, axis=0) & prevMsk).astype(np.uint8)

    ttc_on, ttc_off = ttcDense_int(img.reshape(img.shape[0], -1),
               prevImg.reshape(img.shape[0], -1), _msk)
    s_on = np.einsum( 'ia, ja -> ij', _msk.reshape(_msk.shape[0], -1), img.reshape(img.shape[0], -1), 
                                optimize='optimal', dtype=np.int32)
    s_cnt = _msk.sum((1,2))
#   all_ttcs_on.append(ttc_on)
#   all_ttcs_off.append(ttc_off)
    all_ttcs_on =  np.zeros( [256] + list(ttc_on.shape), dtype=img.dtype )
    all_ttcs_off=  np.zeros( [256] + list(ttc_off.shape), dtype=img.dtype )
    all_s_on =  np.zeros( [256] + list(ttc_on.shape)[:2], dtype=img.dtype )
    all_s_cnt=  np.zeros( [256, ttc_on.shape[0]], dtype=img.dtype )

    all_ttcs_on[j]  = ttc_on.copy()
    all_ttcs_off[j] = ttc_off.copy()
    all_s_on[j] = s_on.copy()
    all_s_cnt[j] = s_cnt.copy()

    prevImg[:] = img
    prevMsk[:] = ~np.any(cmsk, axis=0)
    
    for tid, img, cmsk in iterator:
        j += 1
        tids.append(tid)

        _msk = (m & ~np.any(cmsk, axis=0) & prevMsk).astype(np.uint8)
        ttc_on, ttc_off = ttcDense_int(img.reshape(img.shape[0], -1),
                prevImg.reshape(img.shape[0], -1), _msk)
        s_on = np.einsum( 'ia, ja -> ij', _msk.reshape(_msk.shape[0], -1), img.reshape(img.shape[0], -1), 
                                    optimize='optimal', dtype=np.int32)
        s_cnt = _msk.sum((1,2))
    
        all_ttcs_on[j]  = ttc_on.astype(np.float32).copy()
        all_ttcs_off[j] = ttc_off.astype(np.float32).copy()
        all_s_on[j] = s_on.copy()
        all_s_cnt[j] = s_cnt.copy()
        #print(tid, j, flush=True)
        prevImg[:] = img
        prevMsk[:] = ~np.any(cmsk, axis=0)
    
    dx, dy = np.diag_indices(img.shape[0])
    all_ttcs_on = all_ttcs_on + np.swapaxes(all_ttcs_on,-2,-1)
    all_ttcs_on[:,:, dx, dy] //= 2
    all_ttcs_off = all_ttcs_off + np.swapaxes(all_ttcs_off,-2,-1)
    all_ttcs_off[:,:, dx, dy] //= 2

    return tids, all_ttcs_on[:j+1], all_ttcs_off[:j+1], all_s_on[:j+1], all_s_cnt[:j+1]

##############
# SPARSE TTC
##############

@dataclass
class TTCdata:
    user_q_mask = None
    current_image = None
    current_image_sparse = None
    current_cal_mask = None
    prev_image = None
    prev_image_sparse = None
    prev_mask = None
    ttc_on = None
    ttc_off = None
    s_on = None
    s_off = None
    s_cnt = None

def do_sparse_train(data: TTCdata):
    if data.prev_image is None:
        data.prev_image = np.ones_like(data.current_image)
        data.prev_image_sparse = data.current_image_sparse # sparsify(data.prev_image.reshape(data.prev_image.shape[0], -1))
        data.prev_mask = np.ones_like(data.user_q_mask)

    start = time.time()
    _msk = (data.user_q_mask & ~np.any(data.current_cal_mask, axis=0) & data.prev_mask).astype(np.uint8)
    # print(f"Mask: {time.time() - start:.3f}")

    _im = data.current_image.reshape(data.current_image.shape[0], -1)
    start = time.time()
    a = data.current_image_sparse # sparsify(_im)
    # print(f"Sparsifying ({a.shape}): {time.time() - start:.3f}")
    reps =  (_im > 0).sum(0).cumsum().astype(np.uint32)

    _imPrev = data.prev_image.reshape(data.prev_image.shape[0], -1)
    b = data.prev_image_sparse # sparsify(_imPrev)
    start = time.time()
    repsPrev =  (_imPrev > 0).sum(0).cumsum().astype(np.uint32)
    # print(f"repsPrev: {time.time() - start:.3f}")

    start = time.time()
    data.s_cnt = _msk.sum((1,2))
    # print(f"s_cnt: {data.s_cnt}")
    ttc_on, ttc_off, s_on, s_off  = twoTime(a, reps, b, repsPrev,
                                            _msk, data.current_image.shape[0])
    # print(f"TTCF: {time.time() - start:.3f}")
    data.ttc_on = ttc_on
    data.ttc_off = ttc_off
    data.s_on = s_on
    data.s_off = s_off

    data.prev_image = data.current_image
    data.prev_image_sparse = data.current_image_sparse # a
    data.prev_mask = ~np.any(data.current_cal_mask, axis=0)

def do_sparse(iterator, m):
    
#   all_ttcs = list()

    tids = list()

    tid, img, cmsk = iterator.__next__()
    tids.append(tid)
    j = 0

    ttc_data = TTCdata()
    ttc_data.user_q_mask = m
    ttc_data.current_image = img
    ttc_data.current_cal_mask = cmsk

    do_sparse_train(ttc_data)

    ttc_on_shape = list(ttc_data.ttc_on.shape)
    all_ttcs_on =  np.zeros( [256] + ttc_on_shape, dtype=img.dtype )
    all_ttcs_off =  np.zeros( [256] + ttc_on_shape, dtype=img.dtype )
    all_s_on = np.zeros( [256] + ttc_on_shape[:2], dtype=img.dtype )
    all_s_off = np.zeros( [256] + ttc_on_shape[:2], dtype=img.dtype )
    all_s_cnt = np.zeros( [256, ttc_on_shape[0]], dtype=img.dtype )

    all_ttcs_on[j]  = ttc_data.ttc_on.copy()
    all_ttcs_off[j]  = ttc_data.ttc_off.copy()
    all_s_on[j]  = ttc_data.s_on.copy()
    all_s_off[j]  = ttc_data.s_off.copy()
    all_s_cnt[j]  = ttc_data.s_cnt.copy()
    
    for tid, img, cmsk in iterator:
        j += 1
        tids.append(tid)

        ttc_data.current_image = img
        ttc_data.current_cal_mask = cmsk

        do_sparse_train(ttc_data)

        all_ttcs_on[j]  = ttc_data.ttc_on.copy()
        all_ttcs_off[j]  = ttc_data.ttc_off.copy()
        all_s_on[j]  = ttc_data.s_on.copy()
        all_s_off[j]  = ttc_data.s_off.copy()
        all_s_cnt[j]  = ttc_data.s_cnt.copy()
        
    dx, dy = np.diag_indices(img.shape[0])
    all_ttcs_on = all_ttcs_on + np.swapaxes(all_ttcs_on,-2,-1)
    all_ttcs_on[:,:, dx, dy] //= 2

    return tids, all_ttcs_on[:j+1], all_ttcs_off[:j+1], all_s_on[:j+1], all_s_off[:j+1], all_s_cnt[:j+1]

##############
# NUMPY EINSUM
##############

def do_traditional(iterator, m):
    
    tids = list()
    
    tid, img, cmsk  = iterator.__next__()
    tids.append(tid)
    j = 0

    prevImg = np.ones_like(img, dtype=img.dtype)
    prevCmsk = np.ones_like(cmsk, dtype=img.dtype)

    ttc_on  = np.zeros( (m.shape[0], img.shape[0], img.shape[0]), dtype=img.dtype)
    ttc_off = np.ones_like(ttc_on)
    s_on = np.ones((m.shape[0], img.shape[0], ), dtype=img.dtype)
    s_off = np.ones((m.shape[0], img.shape[0], ), dtype=img.dtype)
    s_cnt = np.ones((m.shape[0]), dtype=img.dtype)

    _ttc = np.zeros( (img.shape[0], img.shape[0]), dtype=img.dtype )

    all_ttcs_on  = np.zeros( (256, m.shape[0], img.shape[0], img.shape[0]), dtype=img.dtype )
    all_ttcs_off = np.ones_like(all_ttcs_on)
    all_s_on = np.ones( (256, m.shape[0], img.shape[0], ), dtype=img.dtype )
    all_s_off = np.ones( (256, m.shape[0], img.shape[0], ), dtype=img.dtype )
    all_s_cnt = np.ones( (256, m.shape[0]), dtype=img.dtype )


    for i, _m in enumerate(m):
        _msk = (_m & ~(np.any(cmsk, axis=0)) & ~(np.any(prevCmsk, axis=0))).astype(_m.dtype)
        _a = img[:, _msk].reshape(img.shape[0], -1)
        _ttc = np.einsum( 'ia, ja -> ij', _a, _a, optimize='optimal', dtype=np.int32)
        _s = np.nansum(_a,1)
        ttc_on[i]  = _ttc 
        s_on[i]  = _s
        s_cnt[i]  = np.nansum(_msk)

        _b = prevImg[:, _msk].reshape(img.shape[0], -1)
        _ttc = np.einsum( 'ia, ja -> ij', _a, _b, optimize='optimal', dtype=np.int32)
        ttc_off[i] = _ttc 
        _s_off = np.nansum(_b,1)
        s_off[i]  = _s_off

    all_ttcs_on[j] = ttc_on.copy()
    all_ttcs_off[j] = ttc_off.copy()
    all_s_on[j] = s_on.copy()
    all_s_off[j] = s_off.copy()
    all_s_cnt[j] = s_cnt.copy()

    prevImg[:] = img

    for tid, img, cmsk in iterator:
        j += 1
        tids.append(tid)
    
        for i, _m in enumerate(m):
            _msk = (_m & ~(np.any(cmsk, axis=0)) & ~(np.any(prevCmsk, axis=0))).astype(_m.dtype)
            _a = img[:, _msk].reshape(img.shape[0], -1)
            _ttc = np.einsum( 'ia, ja -> ij', _a, _a, optimize='optimal', dtype=np.int32)
            _s = np.nansum(_a,1)
            ttc_on[i]  = _ttc 
            s_on[i]  = _s
            s_cnt[i]  = np.nansum(_msk)


            _b = prevImg[:, _msk].reshape(img.shape[0], -1)
            _ttc = np.einsum( 'ia, ja -> ij', _a, _b, optimize='optimal', dtype=np.int32)
            ttc_off[i] = _ttc 
            _s_off = np.nansum(_b,1)
            s_off[i]  = _s_off 

        all_ttcs_on[j] = ttc_on.copy()
        all_ttcs_off[j] = ttc_off.copy()
        all_s_on[j] = s_on.copy()
        all_s_off[j] = s_off.copy()
        all_s_cnt[j] = s_cnt.copy()

        #print(tid, j, flush=True)

        prevImg[:] = img
        prevCmsk[:] = cmsk
        
    return tids, all_ttcs_on[:j+1], all_ttcs_off[:j+1], all_s_on[:j+1], all_s_off[:j+1], all_s_cnt[:j+1]
