cimport cython

import numpy as np
from scipy.linalg.cython_blas cimport ssyr, ssyr2

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef ttcDense(float[:,::1] data, 
               float[:,::1] prevDat, 
               char[:,:,::1]  mask):
    
    cdef int i, j, k, m
    
    cdef int dim = data.shape[0]
    cdef int n_reg = mask.shape[0]
    

#   cdef int n_trn = data.shape[0]
    cdef int n_pix = data.shape[1]
    
    cdef float[:,:,::1] A = np.zeros(( n_reg, dim, dim), dtype=np.float32)
    cdef float[:,:,::1] B = np.zeros_like(A)
     
    cdef int one = 1
    cdef float fac = 1., half=0.5

    cdef int[::1]  ixmsk = np.argmax(mask, axis=0).ravel().astype(np.int32)
    cdef char[::1] cond = np.any(mask, axis=0).ravel().astype(np.int8)
    
#   for i in range(n_trn):
    for j in range(n_pix):
        
        if cond[j]:
            m = ixmsk[j]

            ssyr(
            "L",
            &dim,
            &fac,
            &data[0 , j],
            &n_pix,
            &A[m,0,0],
            &dim
            )

            ssyr2(
            "L",
            &dim,
            &half,
            &data[ 0 , j],
            &n_pix,
            &prevDat[ 0 , j],
            &n_pix,
            &B[m,0,0],
            &dim
            )
                
    
    return np.asarray(A), np.asarray(B)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef _ttc_upd(int val,
              int* data, 
              int* prevDat, 
              int* A,
              int* B,
              int size,
              int stride):

    cdef int l, v2

    for l in range(size):
        v2 = data[l * stride]
        A[l] += val * v2

        v2 = prevDat[l * stride]
        B[l] += val * v2

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef ttcDense_int(int[:,::1] data, 
                   int[:,::1] prevDat, 
                   char[:,:,::1]  mask):
    
    cdef int i, j, k, m, l
    cdef int v1, v2
    
    cdef int dim = data.shape[0]
    cdef int n_reg = mask.shape[0]
    

#   cdef int n_trn = data.shape[0]
    cdef int n_pix = data.shape[1]
    
    cdef int[:,:,::1] A = np.zeros(( n_reg, dim, dim), dtype=np.int32)
    cdef int[:,:,::1] B = np.zeros_like(A)
     
    cdef int[::1]  ixmsk = np.argmax(mask, axis=0).ravel().astype(np.int32)
    cdef char[::1] cond = np.any(mask, axis=0).ravel().astype(np.int8)
    
    for j in range(n_pix):
        
        if cond[j]:
            m = ixmsk[j]

            for k in range(dim):
                v1 = data[k, j]

                if v1:
                    _ttc_upd(v1,
                            &data[0,j],
                            &prevDat[0,j],
                            &A[m,k,0],
                            &B[m,k,0],
                            dim,
                            n_pix)


    
    return np.asarray(A), np.asarray(B)




@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef ttcDenseSymmNorm(float[:,:,::1] data, 
                       char[:,:,::1] mask):
    
    cdef int i, j, k, l, m
    
    cdef int dim = data.shape[1]
    cdef int n_reg = mask.shape[0]
    

    cdef int n_trn = data.shape[0]
    cdef int n_pix = data.shape[2]
    
    cdef float[:,:,:,::1] A = np.zeros((n_trn, n_reg, dim, dim), dtype=np.float32)
    cdef float[:,::1]    _a = np.zeros((dim, dim),               dtype=np.float32 )
    cdef float[:,:,::1]   N = np.zeros((n_trn, n_reg, dim),      dtype=np.float32)
     
    cdef int one = 1
    cdef float fac = 1.
    cdef float cs_left, cs_rght, nrm

    cdef int[::1] ixmsk = np.argmax(mask, axis=0).ravel().astype(np.int32)
    cdef char[::1] cond = np.any(mask, axis=0).ravel().astype(np.int8)
    
    for i in range(n_trn):
        for j in range(n_pix):
            
            if cond[j]:
                m = ixmsk[j]
                
                _a[:] = 0

                ssyr(
                "L",
                &dim,
                &fac,
                &data[i, 0 , j],
                &n_pix,
                &_a[0,0],
                &dim
                )
                
                cs_left = 0
                cs_rght = 0
                for k in range(dim):
                    cs_left += data[i, k,     j]
                    cs_rght += data[i, dim-k-1, j]
                    
                    if cs_left and cs_rght:
                        nrm =  ((k+1) * (k+1)) / (cs_left * cs_rght)
                    else:
                        nrm = 0
                    
                    for l in range(k+1):
                        
                        A[i,m,l,dim-k+l-1] += nrm *  _a[l, dim-k+l-1]
    
    return np.asarray(A)
