cimport cython
import numpy as np

@cython.boundscheck(False)
@cython.wraparound(False)
def sparsify( int[:,::1] data):
    
    cdef int k = data.shape[0]
    cdef int l = data.shape[1]
    
    cdef int i, j
    
    cdef list indices = []
    
    cdef int a
    cdef unsigned int b
    
    ix_old = -1
    
    for j in range(l):
        for i in range(k):

            a = data[i,j]
            if a:
                b  = 0           #    16 bit         9 bit     7 bit
                b |= (j << 16)   ####pixel#######  ###cell##  #count#
                b |= (i << 7)
                
                b |=  (a & 0x7F)
                
                indices.append(b)
                
                
                
    return np.asarray(indices, dtype=np.uint32) 

@cython.wraparound(False)
def sparsify2(unsigned int[::1] pixels, unsigned int[:, ::1] indices, unsigned int[::1] out):
    cdef unsigned int elem, cell, idx

    for i in range(pixels.shape[0]):
        cell = indices[1, i]
        idx = indices[0, i]

        elem = 0
        elem |= (idx << 16)
        elem |= (cell << 7)
        elem |= (pixels[i] & 0x7F)
        out[i] = elem

@cython.boundscheck(False)
@cython.wraparound(False)
def densify( unsigned int[::1] indices,
             tuple  size):
    cdef int i, j
    
    cdef short[:,::1] A = np.zeros(size, dtype=np.int16)
    
    cdef int col,  val
    cdef int  row
    
    for i in range(len(indices)):
        val  = indices[i]
        row  = (val >> 7) & 0x1FF
        col = val >> 16
        val &= 0x7F

        A[row, col] = val
        
    return np.asarray(A)
            
    
# @cython.boundscheck(False)
@cython.wraparound(False)
def twoTime( unsigned int[::1] indices,
             unsigned int[::1] reps,
             unsigned int[::1] indices_prev,
             unsigned int[::1] reps_prev,
             char[:,:,::1] mask,
             int dim):
    cdef unsigned int i=0, j=0, k=0
    
    cdef unsigned int n_reg = mask.shape[0]
    cdef int[:,:,::1] ttc_on  = np.zeros( (n_reg, dim, dim), dtype=np.int32)
    cdef int[:,:,::1] ttc_off = np.zeros( (n_reg, dim, dim), dtype=np.int32)
    cdef int[:,::1]   s_on  = np.zeros( (n_reg, dim, ), dtype=np.int32)
    cdef int[:,::1]   s_off = np.zeros( (n_reg, dim, ), dtype=np.int32)
    cdef int[::1]     s_cnt = np.zeros( (n_reg, ), dtype=np.int32)
    
    cdef unsigned int col,  val, v, rmax, rmin, lst_r=0
    cdef unsigned int rP, lst_rP = 0
    cdef unsigned int  row, row2, val2, m

    cdef int[::1] ixmsk = np.argmax(mask, axis=0).ravel().astype(np.int32)
    cdef char[::1] cond = np.any(mask, axis=0).ravel().astype(np.int8)
    
    
    
    for j in range(reps.shape[0]):
        r = reps[j]
        rmax = 0 if r ==0 else r - 1
        rmin = min(lst_r, rmax)
        
        val = indices[rmin]
        
        col = val >> 16

        # prevImg
        rP = reps_prev[j]

#       rmax = 0 if rP ==0 else rP - 1
#       rmin = min(lst_rP, rmax)

#       val = indices_prev[rmin]
#       if r and rP:
#           assert col == (val >> 16)

        if cond[col]:
            m = ixmsk[col]
            i = lst_r
            while i < r:
                val = indices[i]
                row  = (val >> 7) & 0x1FF
                val &= 0x7F

                s_on[m,row] += val

                k = i
                while k <r:
                    val2 = indices[k]
                    row2  = (val2 >> 7) & 0x1FF
                    val2 &= 0x7F

                    ttc_on[m, row, row2] += val * val2

                    k += 1

                # prevImg
                k = lst_rP
                while False: # k < rP:
                    val2 = indices_prev[k]
                    if (val2 >> 16) == col:
                        row2  = (val2 >> 7) & 0x1FF
                        val2 &= 0x7F

                        v = val * val2
                        ttc_off[m, row, row2] += v
#                       ttc_off[m, row2, row] += v

                    k += 1

                        
                i += 1

            i = lst_rP
            while False: # i < rP:
                val2 = indices_prev[i]
                row2  = (val2 >> 7) & 0x1FF
                val2 &= 0x7F
                s_off[m,row2] += val2

                i += 1


        lst_r = r
        lst_rP = rP
            
        
    return np.asarray(ttc_on), np.asarray(ttc_off), np.asarray(s_on), np.asarray(s_off)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def twoTimeSymmNorm( unsigned int[::1] indices,
                     unsigned int[::1] reps,
                     char[:,:,::1] mask,
                     float[:,::1] avg,
                     int dim):
    cdef unsigned int i=0, j=0, k=0, k1, k2, m
    
    cdef int n_reg = mask.shape[0]
    cdef float[:,:,::1] ttc = np.zeros( (n_reg, dim, dim), dtype=np.float32)
    cdef float[::1]   nrm = np.zeros( dim, dtype=np.float32 ) 
    
    
    cdef int[::1] ixmsk = np.argmax(mask, axis=0).ravel().astype(np.int32)
    cdef char[::1] cond = np.any(mask, axis=0).ravel().astype(np.int8)
    
    
    cdef unsigned int col,  val1, val2, r, rmax, rmin, lst_r = 0, col1, col2
    cdef unsigned int  row1, row2
    
    cdef unsigned int j1, j2
    
    cdef float v, s1, s2, n
    
    
    
    for j in range(reps.shape[0]):
        r = reps[j]
        rmax = 0 if r ==0 else r - 1
        rmin = min(lst_r, rmax)
        
        k1 = rmin
        k2 = rmax
        
        val1 = indices[k1]
        row1  = (val1 >> 7) & 0x1FF
        val2 = indices[k2]
        row2  = (val2 >> 7) & 0x1FF
        
        s1 = 0
        s2 = 0
        
        col = val1 >> 16
        
        if cond[col]:
            m = ixmsk[col]
            for i in range(dim):
                
#                 if i == row1:
#                     v = val1 & 0x7F
#                     k1 = min(k1+1, rmax)
#                     val1 = indices[k1]
#                     row1  = (val1 >> 7) & 0x1FF
#                 else: v = 0
                v = avg[i, col]
                s1 += (v-s1)/(i+1)    

#                 if (dim-1-i) == row2:
#                     v = val2 & 0x7F
#                     k2 = max(rmin, k2-1)
#                     val2 = indices[k2]
#                     row2  = (val2 >> 7) & 0x1FF
#                 else: v = 0

                v = avg[dim-1-i, col]

                s2 += (v-s2)/(i+1)

                n = s1 * s2
                nrm[dim-1-i] = 1/n if n > 0 else 0 

            i = lst_r

            while i < r:
                val1 = indices[i]
                col1 = val1 >> 16
                row1  = (val1 >> 7) & 0x1FF
                val1 &= 0x7F
                k = i
                while k <r:
                    val2 = indices[k]
                    col2 = val2 >> 16
                    row2  = (val2 >> 7) & 0x1FF
                    val2 &= 0x7F
                    
                    n = nrm[ row2 - row1 ] 
                    ttc[m, row1, row2] += val1 * val2 * n

                    k += 1

                i += 1
            
            
        lst_r = r
        
    return np.asarray(ttc)
