import time
from pathlib import Path

import numpy as np
from karabo.bound import NODE_ELEMENT, NDARRAY_ELEMENT, STRING_ELEMENT

from fastXPCS.algos import TTCdata, do_sparse_train
from fastXPCS.fxpcs_sparse import sparsify, sparsify2
from calng.correction_addons.base_addon import BaseCorrectionAddon


class Xpcs(BaseCorrectionAddon):
    def __init__(self, config):
        self._ttcdata = TTCdata()
        self._mask_path = config["mask_path"]
        self._module_no = None

        self._current_image_mean = 0
        self._current_image = None
        self._sparse_array = None
        self._cal_mask = None

        global cupy
        import cupy

    @staticmethod
    def extend_device_schema(schema, prefix):
        (
            STRING_ELEMENT(schema)
            .key(f"{prefix}.mask_path")
            .tags("managed")
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),
        )

    @staticmethod
    def extend_output_schema(schema):
        (
            NODE_ELEMENT(schema)
            .key("xpcs")
            .commit(),

            NDARRAY_ELEMENT(schema)
            .key("xpcs.ttcf_on")
            .dtype('FLOAT')
            .commit(),

            NDARRAY_ELEMENT(schema)
            .key("xpcs.ttcf_off")
            .dtype('FLOAT')
            .commit(),
        )

    def set_mask(self):
        if Path(self._mask_path).is_file() and self._module_no is not None:
            mask = np.load(self._mask_path)
            self._ttcdata.user_q_mask = mask[:, self._module_no].copy()
            print(f"Loaded mask for module {self._module_no} with shape {self._ttcdata.user_q_mask.shape}, mean {self._ttcdata.user_q_mask.mean()}")

    def reconfigure(self, changed_config):
        if changed_config.has("mask_path"):
            self._mask_path = changed_config["mask_path"]
            self.set_mask()

    def post_correction(self, data, cell_table, pulse_table, output_hash):
        global cupy

        # Insert dummy values
        data[0] = 0

        # Photonize
        # print(f"Unphotonized mean ({data.shape}): {np.nanmean(data)}")
        start = time.time()
        data /= 10
        data[data < 0] = 0
        np.around(data, out=data)
        # print(f"Photonized mean: {np.nanmean(data)}")

        # Sparsify
        start = time.time()
        flat_image_shape = data.reshape(data.shape[0], -1).shape
        reshaped_data = data.reshape(flat_image_shape).T
        indices = cupy.indices(reshaped_data.shape)

        lit_mask = reshaped_data > 0
        lit_pixels = reshaped_data[lit_mask].astype(cupy.uint32)
        lit_indices = indices[:, lit_mask].astype(cupy.uint32)

        lit_pixels_cpu = lit_pixels.get()
        lit_indices_cpu = lit_indices.get()

        sparse_array = np.empty_like(lit_pixels_cpu, dtype=np.uint32)
        sparsify2(lit_pixels_cpu, lit_indices_cpu, sparse_array)

        self._current_image = data.astype(cupy.int32).get()
        self._current_image_mean = np.nanmean(data)
        self._cal_mask = np.isnan(data).get()
        self._sparse_array = sparse_array
        # print(f"Sparsified in {time.time() - start:.3f} down to: {self._sparse_array.shape}")

        if self._module_no == 3:
            print(f"Found lit pixels: {sparse_array.shape}")

    def post_reshape(self, data, cell_table, pulse_table, output_hash):
        if self._module_no is None:
            source = list(self._device.sources)[0]
            self._module_no = int(source.split("/")[-1].split(":")[0][:-3])
            self.set_mask()

        if self._module_no != 3 or self._current_image_mean > 20:
            return
        
        # print(f"Processing with mean {self._current_image_mean:.3e} and lit pixels: {self._sparse_array.shape}")

        sparse_ref = sparsify(self._current_image.reshape(self._current_image.shape[0], -1))
        # print(sparse_ref.shape, self._sparse_array.shape)
        # if sparse_ref.shape == self._sparse_array.shape:
        #     print(np.all(np.equal(sparse_ref, self._sparse_array)))

        # print(f"Using lit pixels: {self._sparse_array.shape}")
        start = time.time()
        self._ttcdata.current_image = self._current_image
        self._ttcdata.current_image_sparse = sparse_ref # self._sparse_array
        self._ttcdata.current_cal_mask = self._cal_mask

        elapsed = time.time() - start
        # print(f"Init in: {elapsed:.3f}")

        start = time.time()
        do_sparse_train(self._ttcdata)
        elapsed = time.time() - start
        # print(f"TTCF in: {elapsed:.3f}")

        print(np.nanmean(self._ttcdata.ttc_on), self._ttcdata.s_cnt)

        output_hash["xpcs.ttcf_on"] = self._ttcdata.ttc_on.astype(np.int32)
        # output_hash["xpcs.ttcf_off"] = self._ttcdata.ttc_off.astype(np.int32)
